;; Generated package description from bufferlo.el  -*- no-byte-compile: t -*-
(define-package "bufferlo" "0.2" "Manage frame/tab-local buffer lists" '((emacs "27.1")) :commit "931b9ee3462164d6541d3b6078c46eb75f962592" :authors '(("Florian Rommel" . "mail@florommel.de")) :maintainer '("Florian Rommel" . "mail@florommel.de") :keywords '("buffer" "frame" "tabs" "local") :url "https://github.com/florommel/bufferlo")
